import polars as pl
import polars.selectors as cs


def match_hits(
    blastout: pl.LazyFrame, adapter_df: pl.LazyFrame, end_length: int
) -> pl.LazyFrame:
    """Determine whether each blast hit matches sufficient criteria for discard/trimming

    Keyword arguments:
    blastout: pl.DataFrame containing the results of a blast search with read lengths in column 13
    adapter_key: pl.DataFrame of adapters and their filtering/trimming parameters
    end_length: length of window at either end of read which is examined for trimming
    """
    ## When to discard a read due to adapter presence
    discard = (
        (pl.col("discard_middle")).and_(
            (pl.col("read_length") > 2 * end_length),
            (pl.col("qstart") > end_length),
            (pl.col("qend") < pl.col("read_length") - end_length),
            (pl.col("qend") > end_length),
            (pl.col("qstart") < pl.col("read_length") - end_length),
            (pl.col("pident") >= pl.col("middle_pident")),
            (pl.col("length") >= pl.col("middle_length")),
        )
    ) | (
        (pl.col("discard_end")).and_(
            (pl.col("qend") < end_length),
            (pl.col("qstart") > pl.col("read_length") - end_length),
            (pl.col("pident") >= pl.col("end_pident")),
            (pl.col("length") >= pl.col("end_length")),
        )
    )

    ## When to trim a read at the left end
    trim_l = (pl.col("trim_end")).and_(
        pl.col("qend") < end_length,
        (pl.col("pident") >= pl.col("end_pident")),
        (pl.col("length") >= pl.col("end_length")),
    )

    ## When to trim a read at the righr end
    trim_r = (pl.col("trim_end")).and_(
        pl.col("qstart") >= pl.col("read_length") - end_length,
        (pl.col("pident") >= pl.col("end_pident")),
        (pl.col("length") >= pl.col("end_length")),
    )

    return (
        blastout
        ## match each blast hit with an adapter
        .join_where(
            adapter_df, pl.col("sseqid").cast(pl.String).str.contains(pl.col("adapter"))
        )
        ## Determine which hits to keep and return rows with a "real" hit
        .with_columns(discard=discard, trim_l=trim_l, trim_r=trim_r)
        .filter(pl.any_horizontal("discard", "trim_l", "trim_r"))
    )


def determine_actions(
    hits: pl.LazyFrame, end_length: int, min_length: int
) -> pl.LazyFrame:
    """Take a dataframe of matches from match_hits()
    and process the result to decide which actions to take.

    Keyword arguments:
    hits: pl.DataFrame of blast hits with matches
    end_length: length of window at either end of read which is examined for trimming
    min_length: minimum length of a read after trimming to keep
    """
    hits_actions = hits.with_columns(
        (
            pl.col("read_length")
            - ((pl.col("trim_l").any() + pl.col("trim_r").any()) * end_length)
        )
        .over("qseqid")
        .alias("read_length_after_trimming"),
    ).with_columns(
        gen_action=(
            pl.when(
                pl.col("discard").any()
                | (pl.col("read_length_after_trimming") < min_length)
            )
            .then(pl.lit("discard"))
            .otherwise(pl.lit("trim"))
            .over("qseqid")
        )
    )

    discard = (
        hits_actions.filter(pl.col("gen_action") == "discard")
        .group_by("qseqid")
        .agg(
            pl.col("sseqid", "qstart", "qend", "read_length").get(
                pl.col("evalue").arg_min()
            ),
            action=pl.lit("discard"),
        )
    )

    trim_l = (
        hits_actions.filter(~(pl.col("gen_action") == "discard"), pl.col("trim_l"))
        .group_by("qseqid")
        .agg(
            pl.col("sseqid", "qstart", "qend", "read_length")
            .filter(pl.col("qend") <= end_length)
            .get(pl.col("evalue").filter(pl.col("qend") <= end_length).arg_min()),
            action=pl.lit("trim_l"),
        )
    )

    trim_r = (
        hits_actions.filter(~(pl.col("gen_action") == "discard"), pl.col("trim_r"))
        .group_by("qseqid")
        .agg(
            pl.col("sseqid", "qstart", "qend", "read_length")
            .filter(pl.col("qstart") >= pl.col("read_length") - end_length)
            .get(
                pl.col("evalue")
                .filter(pl.col("qstart") >= pl.col("read_length") - end_length)
                .arg_min()
            ),
            action=pl.lit("trim_r"),
        )
    )

    return pl.concat([discard, trim_l, trim_r])



import polars as pl
import polars.selectors as cs


def summarise_blast(blast: pl.LazyFrame) -> pl.DataFrame:
    """
    Summarise a raw BLAST table, returning the number of hits by
    adapter.
    """
    return (
        blast.group_by("sseqid")
        .len("n_hits")
        .rename({"sseqid": "adapter"})
        .sort("adapter")
        .collect()
    )


def summarise_hits(hits: pl.DataFrame) -> pl.DataFrame:
    """
    Summarise a filtered BLAST table, returning the number of hits by
    adapter and action.
    """
    return (
        hits.select(
            sseqid=pl.col("sseqid"),
            discard=pl.col("discard"),
            trim=(pl.col("trim_l") + pl.col("trim_r")).cast(pl.Boolean),
        )
        .unpivot(
            cs.by_name(["discard", "trim"]),
            index="sseqid",
            variable_name="action",
        )
        .filter(pl.col("value"))
        .group_by(["sseqid", "action"])
        .len(name="n_hits")
        .filter(pl.col("n_hits") > 0)
        .rename({"sseqid": "adapter"})
        .sort(["adapter", "action"])
    )


def summarise_actions(actions: pl.DataFrame) -> pl.DataFrame:
    """
    Summarise an actions table, returning the number of reads affected
    and the number of bases removed for each adapter and action.
    """
    return (
        actions.with_columns(
            bases=pl.when(pl.col("action") == "discard")
            .then(pl.col("read_length"))
            .when(pl.col("action").str.contains("trim"))
            .then(pl.lit(150)),
            action=pl.when(pl.col("action") == "discard")
            .then(pl.lit("discard"))
            .otherwise(pl.lit("trim")),
        )
        .group_by(["sseqid", "action"])
        .agg(
            n_reads=pl.col("qseqid").unique().len(), bases_removed=pl.col("bases").sum()
        )
        .filter(pl.col("n_reads") > 0)
        .rename({"sseqid": "adapter"})
        .sort(["adapter", "action"])
    )



import polars as pl
import yaml


def read_adapter_yaml(yaml_path: str) -> pl.LazyFrame:
    """Open an adapter YAML file and return it as a lazy pl.DataFrame"""
    try:
        adapters = yaml.safe_load(yaml_path)
    except yaml.YAMLError as exc:
        print(exc)

    return pl.DataFrame(
        adapters,
        schema={
            "adapter": pl.String,
            "discard_middle": pl.Boolean,
            "discard_end": pl.Boolean,
            "trim_end": pl.Boolean,
            "middle_pident": pl.Float32,
            "end_pident": pl.Float32,
            "middle_length": pl.UInt8,
            "end_length": pl.UInt8,
        },
    ).lazy()


def read_blast(blast_path: str) -> pl.LazyFrame:
    """Read BLAST outformat file to a LazyFrame.

    If a BAM file is provided and there is no 13th column (containing read lengths)
    then these can be quickly counted.

    Returns with the columns qseqid, sseqid, pident, length, qstart, qend, and evalue.
    qstart and qend are sorted to ignore information about strand.
    """
    blastout = pl.scan_csv(
        blast_path,
        has_header=False,
        separator="\t",
        schema={
            "qseqid": pl.String,
            "sseqid": pl.String,
            "pident": pl.Float32,
            "length": pl.UInt8,
            "mismatch": pl.UInt8,
            "gapopen": pl.UInt8,
            "qstart": pl.UInt32,
            "qend": pl.UInt32,
            "sstart": pl.UInt8,
            "send": pl.UInt8,
            "evalue": pl.Float32,
            "bitscore": pl.Float32,
            "read_length": pl.UInt32,
        },
    ).with_columns(
        pl.col("qseqid")
        .cast(pl.Categorical)
        .set_sorted(),  ## Set sorted for faster grouping operations
        pl.col("sseqid").cast(pl.Categorical),
        pl.when(pl.col("qstart") > pl.col("qend"))
        .then(
            pl.struct(
                start="qend",
                end="qstart",
            )
        )
        .otherwise(pl.struct(["qstart", "qend"]))
        .struct.field(["qstart", "qend"]),
    )

    return blastout



import bgzip
import click
import importlib.metadata
import json
import polars as pl
import re


from hifi_trimmer.output import (
    create_bed,
    filter_bam_with_bed,
    write_summary,
    write_hits,
)
from hifi_trimmer.blast import match_hits, determine_actions
from hifi_trimmer.utils import check_records
from hifi_trimmer.read_files import read_adapter_yaml, read_blast
from hifi_trimmer.summary import summarise_blast, summarise_hits, summarise_actions


@click.group()
@click.version_option(package_name="hifi_trimmer")
def cli():
    """Main entry point for the tool."""
    pass


@click.command("process_blast")
@click.argument("blastout", type=click.Path(exists=True))
@click.argument("adapter_yaml", type=click.File(mode="r"))
@click.option(
    "-p",
    "--prefix",
    default=None,
    required=False,
    type=str,
    help="Output prefix for results. Defaults to the basename of the blastout if not provided.",
)
@click.option(
    "-ml",
    "--min_length_after_trimming",
    default=300,
    show_default=True,
    type=int,
    help="Minumum length of a read after trimming the ends in order not to be discarded",
)
@click.option(
    "-el",
    "--end_length",
    default=150,
    show_default=True,
    type=int,
    help="Window size at either end of the read to be considered as 'ends' for searching",
)
@click.option(
    "-hf",
    "--hits",
    "hits_flag",
    default=False,
    is_flag=True,
    type=bool,
    help="""Write the hits identified using the given adapter specifications to TSV. The format is
    standard BLAST outfmt 6 with the following extra columns: read_length (int), discard (bool), trim_l (bool), trim_r (bool)
    """,
)
@click.option(
    "--no-summary",
    default=False,
    type=bool,
    is_flag=True,
    help="Skip writing a summary TSV with the number of hits for each adapter",
)
@click.option(
    "-t",
    "--threads",
    default=1,
    show_default=True,
    type=int,
    help="Number of threads to use for compression",
)
def process_blast(
    blastout: str,
    adapter_yaml: click.File,
    prefix: str,
    min_length_after_trimming: int,
    end_length: int,
    hits_flag: bool,
    no_summary: bool,
    threads: int,
) -> None:
    """Processes the input blastout file according to the adapter yaml key.

    BLASTOUT: tabular file resulting from a BLAST query of a readset against a BLAST
    database of adapter sequences, run with -outfmt "6 std qlen". If the qlen column
    is missing, lengths can be calculated by passing the --bam option.

    ADAPTER_YAML: yaml file containing a list with the following fields per adapter:

    \b
    - name: (name of adapter. can be a regular expression)
      discard_middle: True/False (discard read if adapter found in middle)
      discard_end: True/False (discard read if adapter found in end)
      trim_end: True/False (trim read if adapter found in end)
      middle_pident: int (minimum pident requred to identify adapter in middle of read)
      middle_length: int (minimum match length required to identify adapter in middle of read)
      end_pident: int (minimum pident requred to identify adapter in end window)
      end_length: int (minimum match length requred to identify adapter in end window)

    Output:
    By default, writes bgzipped BED to [prefix].bed.gz, and a JSON summary file with raw counts
    of adapter hits detected, counts identified after processing, and the total length of
    removed sequences per adapter to [prefix].summary.json.
    """
    if prefix is None:
        prefix = re.sub("\\.blastout.*$", "", blastout)

    ## Set null outputs
    blast_summary = None
    hits_summary = None
    actions_summary = None
    out_bed = "".encode("utf-8")

    try:
        blast = read_blast(blastout)
        blast_summary = summarise_blast(blast)
        adapters = read_adapter_yaml(adapter_yaml)

        ## Make sure each barcode that matches an adapter matches only one adapter
        check = check_records(blast, adapters)
        if len(check) > 0:
            raise click.ClickException(
                f"Barcodes {', '.join(check)} match to more than one adapter in {adapter_yaml.name}!"
            )

        ## process the blastout file
        hits = match_hits(blast, adapters, end_length).collect()
        hits_summary = summarise_hits(hits)

        ## Check if any hits matched - need to handle empty input!
        if not (hits.is_empty()):
            actions = determine_actions(
                hits.lazy(), end_length, min_length_after_trimming
            ).collect()
            actions_summary = summarise_actions(actions)

            bed = create_bed(actions.lazy(), end_length)

            if hits_flag:
                write_hits(hits).write_csv(
                    prefix + ".hits", separator="\t", include_header=False
                )

            out_bed = (
                bed.collect()
                .write_csv(separator="\t", include_header=False)
                .encode("utf-8")
            )

            click.echo(f"BLAST file {blastout} parsed successfully!")
        else:
            click.echo(
                "WARNING: After filtering, no BLAST hits remain. Writing empty BED output."
            )
    except pl.exceptions.NoDataError:
        click.echo(f"WARNING: {blastout} was empty! Writing empty BED output.")
        blast = None

    if not no_summary:
        with open(prefix + ".summary.json", "w") as f:
            summary = write_summary(blast_summary, hits_summary, actions_summary)
            json.dump(summary, f, indent=4)

    with open(prefix + ".bed.gz", "wb") as f:
        with bgzip.BGZipWriter(f, num_threads=threads) as out:
            out.write(out_bed)


@click.command("filter_bam")
@click.argument("bam", type=click.File(mode="rb"))
@click.argument("bed", type=click.Path(exists=True))
@click.argument("outfile", default=None, required=True, type=click.File(mode="wb"))
@click.option(
    "-f",
    "--fastq",
    is_flag=True,
    default=False,
    help="Write FASTQ instead of FASTA",
)
@click.option(
    "-t",
    "--threads",
    default=1,
    show_default=True,
    type=int,
    help="Number of threads to use for compression",
)
def filter_bam(
    bam: click.File, bed: str, outfile: click.File, threads: int, fastq: bool
) -> None:
    """
    Filter the reads stored in a BAM file using the appropriate BED file produced
    by blastout_to_bed and write to a bgzipped fasta file.

    \b
    BAM: BAM file in which to filter reads
    BED: BED file describing regions of the read set to exclude.
    OUTFILE: File to write the filtered reads to (bgzipped).
    """
    filters = filter_bam_with_bed(outfile, bam, bed, threads, fastq)

    try:
        read = next(filters)
        print(f"Read {read['read']} not processed!")
        for read in filters:
            print(f"Read {read['read']} not processed!")

        raise click.ClickException(
            "Not all entries in the BED file were processed! Did you forget to sort them in the same order as the BAM file?"
        )
    except StopIteration:
        print("Read filtering complete!")


cli.add_command(process_blast)
cli.add_command(filter_bam)

if __name__ == "__main__":
    cli()



import polars as pl


def check_records(blast: pl.LazyFrame, adapter_df: pl.DataFrame) -> list:
    """Check if any adapters in a BLAST dataframe match to more than one
    adapter sequence in the YAML file.
    """
    counts = (
        blast.select(pl.col("sseqid"))
        .unique()
        .with_columns(pl.col("sseqid").cast(pl.String))
        .join_where(adapter_df, pl.col("sseqid").str.contains(pl.col("adapter")))
        .group_by("sseqid")
        .len()
        .filter(pl.col("len") > 1)
        .collect()["sseqid"]
        .to_list()
    )

    return counts


def format_fastx_record(header: str, sequence: str, qual: str | None) -> str:
    """Format a header and sequence into FASTA format"""
    if qual is not None:
        return f"@{header}\n{sequence}\n+\n{qual}\n"
    else:
        return f">{header}\n{sequence}\n"


def trim_positions(seq: str, ranges: list) -> str:
    """Trim DNA sequence seq to remove the positions specified in ranges.

    seq: string
    ranges: list of non-overlapping tuples with (start, end)
    """
    ## sort the ranges so we trim from the end - that way indexing stays constant
    ranges = sorted(ranges, key=lambda x: x[0], reverse=True)

    for start, end in ranges:
        seq = seq[:start] + seq[end:]

    return seq



import bgzip
import click
import polars as pl
import pysam

from hifi_trimmer.utils import trim_positions, format_fastx_record


def create_bed(actions: pl.LazyFrame, end_length: int) -> pl.LazyFrame:
    """Format the output from determine_actions() into valid BED."""
    return actions.select(
        qseqid=pl.col("qseqid"),
        start=pl.when((pl.col("action") == "discard") | (pl.col("action") == "trim_l"))
        .then(pl.lit(0))
        .when(pl.col("action") == "trim_r")
        .then(pl.col("read_length") - end_length)
        .otherwise(pl.lit(0)),
        end=pl.when((pl.col("action") == "discard") | (pl.col("action") == "trim_r"))
        .then(pl.col("read_length"))
        .when(pl.col("action") == "trim_l")
        .then(pl.lit(end_length))
        .otherwise(pl.lit(0)),
        reason=pl.concat_str(pl.col("action"), pl.lit(":"), pl.col("sseqid")),
    ).sort("qseqid", "start")


def write_summary(
    blast_summary: pl.DataFrame,
    hits_summary: pl.DataFrame,
    actions_summary: pl.LazyFrame,
) -> dict:
    """
    Takes the raw BLAST table, the BLAST table filtered for "valid" hits
    using criteria from the YAML, and the per-read "actions" table,
    and returns a dictionary with a summary about the adapter removal process:

    detected: how many hits per adapter detected in the raw BLAST output
    hits: for each adapter and action, the following stats:
        - number of blast hits
        - number of reads where this was the best hit
        - number of bases removed
    total_reads_discarded: total number of reads discarded
    total_reads_trimmed: total number of reads trimmed
    total_bases_removed: total length of removed bases
    """
    summary = {
        "detections": [],
        "hits": [],
        "total_reads_discarded": 0,
        "total_reads_trimmed": 0,
        "total_bases_removed": 0,
    }

    if blast_summary is not None:
        summary["detections"] = blast_summary.to_dicts()

        if hits_summary is not None and not hits_summary.is_empty():
            summary["hits"] = (
                hits_summary.join(actions_summary, on=["adapter", "action"])
                .sort(["adapter", "action"])
                .to_dicts()
            )

            summary["total_bases_removed"] = actions_summary["bases_removed"].sum()
            summary["total_reads_discarded"] = actions_summary.filter(
                pl.col("action") == "discard"
            )["n_reads"].sum()
            summary["total_reads_trimmed"] = actions_summary.filter(
                pl.col("action") == "trim"
            )["n_reads"].sum()

    return summary


def write_hits(hits: pl.LazyFrame) -> pl.LazyFrame:
    return hits.select(
        [
            "qseqid",
            "sseqid",
            "pident",
            "length",
            "mismatch",
            "gapopen",
            "qstart",
            "qend",
            "sstart",
            "send",
            "evalue",
            "bitscore",
            "read_length",
            "discard",
            "trim_l",
            "trim_r",
        ]
    ).collect()


def filter_bam_with_bed(
    outfile: str, bam: click.File, bed: str, threads: int, fastq: bool
) -> iter:
    try:
        bed_df = pl.read_csv(
            bed,
            separator="\t",
            has_header=False,
            schema={
                "read": pl.String,
                "start": pl.Int64,
                "end": pl.Int64,
                "reason": pl.String,
            },
        )
        filters = bed_df.iter_rows(named=True)

        r = next(filters, None)
    except pl.exceptions.NoDataError:
        click.echo("WARN: BED file is empty! Reads will be streamed as-is.")
        filters = iter([])
        r = next(filters, None)

    with bgzip.BGZipWriter(outfile, num_threads=threads) as out:
        with pysam.AlignmentFile(bam, "rb", check_sq=False, require_index=False) as b:
            ## Process: for each read in the BAM, check if it matches the current BED record.
            ## If yes, pull BED records until we reach a record for the next read.
            ## Then trim the sequence using the records pulled and write to fasta.
            ## If not, write record straight to disk
            for read in b.fetch(until_eof=True):
                if r is not None and r["read"] == read.query_name:
                    ranges = [(int(r["start"]), int(r["end"]))]

                    while True:
                        r = next(filters, None)
                        if r is None or r["read"] != read.query_name:
                            break
                        ranges.append((int(r["start"]), int(r["end"])))

                    sequence = trim_positions(read.query_sequence, ranges)

                    qual = None
                    if fastq:
                        qual = trim_positions(read.qual, ranges)

                    print(
                        f"Processing read: {read.query_name}, ranges: {ranges}, original length: {read.query_length}, new_length: {len(sequence)}"
                    )
                    if len(sequence) > 0:
                        out.write(
                            format_fastx_record(read.query_name, sequence, qual).encode(
                                "utf-8"
                            )
                        )
                else:
                    qual = None
                    if fastq:
                        qual = read.qual

                    out.write(
                        format_fastx_record(
                            read.query_name, read.query_sequence, qual
                        ).encode("utf-8")
                    )

    ## return the remaining filters - if we did all reads, should be empty
    return filters
